@@ -1,9 +1,125 @@
-#SBPS
-#Takes a weightit object and moderator and finds best combination. Smooth version available.
-#Take any two arbitrary weightit objects
-#Take any two propensity score/weights
-#Smooth version only for ps
+#' Subgroup Balancing Propensity Score
+#'
+#' @description
+#' Implements the subgroup balancing propensity score (SBPS), which is an
+#' algorithm that attempts to achieve balance in subgroups by sharing
+#' information from the overall sample and subgroups (Dong, Zhang, Zeng, & Li,
+#' 2020; DZZL). Each subgroup can use either weights estimated using the whole
+#' sample, weights estimated using just that subgroup, or a combination of the
+#' two. The optimal combination is chosen as that which minimizes an imbalance
+#' criterion that includes subgroup as well as overall balance.
+#'
+#' @param obj a `weightit` object containing weights estimated in the
+#' overall sample.
+#' @param obj2 a `weightit` object containing weights estimated in the
+#' subgroups. Typically this has been estimated by including `by` in the
+#' call to [weightit()]. Either `obj2` or `moderator` must be
+#' specified.
+#' @param moderator optional; a string containing the name of the variable in
+#' `data` for which weighting is to be done within subgroups or a
+#' one-sided formula with the subgrouping variable on the right-hand side. This
+#' argument is analogous to the `by` argument in `weightit()`, and in
+#' fact it is passed on to `by`. Either `obj2` or `moderator`
+#' must be specified.
+#' @param formula an optional formula with the covariates for which balance is
+#' to be optimized. If not specified, the formula in `obj$call` will be
+#' used.
+#' @param data an optional data set in the form of a data frame that contains
+#' the variables in `formula` or `moderator`.
+#' @param smooth `logical`; whether the smooth version of the SBPS should
+#' be used. This is only compatible with `weightit` methods that return a
+#' propensity score.
+#' @param full.search `logical`; when `smooth = FALSE`, whether every
+#' combination of subgroup and overall weights should be evaluated. If
+#' `FALSE`, a stochastic search as described in DZZL will be used instead.
+#' If `TRUE`, all \eqn{2^R} combinations will be checked, where \eqn{R} is
+#' the number of subgroups, which can take a long time with many subgroups. If
+#' unspecified, will default to `TRUE` if \eqn{R <= 8} and `FALSE`
+#' otherwise.
+#'
+#' @returns
+#' A `weightit.sbps` object, which inherits from `weightit`.
+#' This contains all the information in `obj` with the weights, propensity
+#' scores, call, and possibly covariates updated from `sbps()`. In
+#' addition, the `prop.subgroup` component contains the values of the
+#' coefficients C for the subgroups (which are either 0 or 1 for the standard
+#' SBPS), and the `moderator` component contains a data.frame with the
+#' moderator.
+#'
+#' This object has its own summary method and is compatible with \pkg{cobalt}
+#' functions. The `cluster` argument should be used with \pkg{cobalt}
+#' functions to accurately reflect the performance of the weights in balancing
+#' the subgroups.
+#'
+#' @details
+#' The SBPS relies on two sets of weights: one estimated in the overall sample
+#' and one estimated within each subgroup. The algorithm decides whether each
+#' subgroup should use the weights estimated in the overall sample or those
+#' estimated in the subgroup. There are 2^R permutations of overall and
+#' subgroup weights, where R is the number of subgroups. The optimal
+#' permutation is chosen as that which minimizes a balance criterion as
+#' described in DZZL. The balance criterion used here is, for binary and
+#' multi-category treatments, the sum of the squared standardized mean differences
+#' within subgroups and overall, which are computed using
+#' [cobalt::col_w_smd()], and for continuous treatments, the
+#' sum of the squared correlations between each covariate and treatment within
+#' subgroups and overall, which are computed using [cobalt::col_w_corr()].
+#'
+#' The smooth version estimates weights that determine the relative
+#' contribution of the overall and subgroup propensity scores to a weighted
+#' average propensity score for each subgroup. If P_O are the propensity scores
+#' estimated in the overall sample and P_S are the propensity scores estimated
+#' in each subgroup, the smooth SBPS finds R coefficients C so that for each
+#' subgroup, the ultimate propensity score is \eqn{C*P_S + (1-C)*P_O}, and
+#' weights are computed from this propensity score. The coefficients are
+#' estimated using [optim()] with `method = "L-BFGS-B"`. When C is
+#' estimated to be 1 or 0 for each subgroup, the smooth SBPS coincides with the
+#' standard SBPS.
+#'
+#' If `obj2` is not specified and `moderator` is, `sbps()` will
+#' attempt to refit the model specified in `obj` with the `moderator`
+#' in the `by` argument. This relies on the environment in which
+#' `obj` was created to be intact and can take some time if `obj` was
+#' hard to fit. It's safer to estimate `obj` and `obj2` (the latter
+#' simply by including the moderator in the `by` argument) and supply
+#' these to `sbps()`.
+#'
+#' @seealso
+#' [weightit()], [summary.weightit()]
+#'
+#' @references
+#' Dong, J., Zhang, J. L., Zeng, S., & Li, F. (2020). Subgroup
+#' balancing propensity score. Statistical Methods in Medical Research, 29(3),
+#' 659–676. \doi{10.1177/0962280219870836}
+#'
+#' @examples
+#'
+#' library("cobalt")
+#' data("lalonde", package = "cobalt")
+#'
+#' #Balancing covariates between treatment groups within races
+#' (W1 <- weightit(treat ~ age + educ + married +
+#'                 nodegree + race + re74, data = lalonde,
+#'                 method = "glm", estimand = "ATT"))
+#'
+#' (W2 <- weightit(treat ~ age + educ + married +
+#'                 nodegree + race + re74, data = lalonde,
+#'                 method = "glm", estimand = "ATT",
+#'                 by = "race"))
+#' S <- sbps(W1, W2)
+#' print(S)
+#' summary(S)
+#' bal.tab(S, cluster = "race")
+#'
+#' #Could also have run
+#' #  sbps(W1, moderator = "race")
+#'
+#' S_ <- sbps(W1, W2, smooth = TRUE)
+#' print(S_)
+#' summary(S_)
+#' bal.tab(S_, cluster = "race")
 
+#' @export
 sbps <- function(obj, obj2 = NULL, moderator = NULL, formula = NULL, data = NULL, smooth = FALSE, full.search) {
 
   if (is_null(obj2) && is_null(moderator)) {
@@ -17,7 +133,7 @@ sbps <- function(obj, obj2 = NULL, moderator = NULL, formula = NULL, data = NULL
   estimand <- obj[["estimand"]]
 
   data.list <- list(data, obj2[["covs"]], obj[["covs"]])
-  combined.data <- do.call(data.frame, clear_null(data.list))
+  combined.data <- do.call("data.frame", clear_null(data.list))
   processed.moderator <- process.by(moderator, data = clear_null(combined.data),
                                     treat = obj[["treat"]], treat.name = NULL,
                                     by.arg = "moderator")
@@ -279,7 +395,7 @@ sbps <- function(obj, obj2 = NULL, moderator = NULL, formula = NULL, data = NULL
     }
 
     if (full.search) {
-      S <- as.matrix(setNames(do.call(expand.grid, replicate(length(R), c(0, 1), simplify = FALSE)),
+      S <- as.matrix(setNames(do.call("expand.grid", replicate(length(R), c(0, 1), simplify = FALSE)),
                               R))
 
       F_min <- Inf
@@ -361,6 +477,7 @@ sbps <- function(obj, obj2 = NULL, moderator = NULL, formula = NULL, data = NULL
   out
 }
 
+#' @exportS3Method print weightit.sbps
 print.weightit.sbps <- function(x, ...) {
   treat.type <- get_treat_type(x[["treat"]])
   trim <- attr(x[["weights"]], "trim")
@@ -391,6 +508,8 @@ print.weightit.sbps <- function(x, ...) {
   }
   invisible(x)
 }
+
+#' @exportS3Method summary weightit.sbps
 summary.weightit.sbps <- function(object, top = 5, ignore.s.weights = FALSE, ...) {
 
   if (ignore.s.weights || is_null(object$s.weights)) sw_ <- rep(1, length(object$weights))
@@ -472,9 +591,6 @@ summary.weightit.sbps <- function(object, top = 5, ignore.s.weights = FALSE, ...
         nn["Weighted", i] <- ESS(w[t==i])
       }
     }
-    else if (treat.type == "ordinal") {
-      .err("Sneaky, sneaky! Ordinal coming soon :)", tidy = FALSE)
-    }
 
     out$effective.sample.size <- nn
 
@@ -483,7 +599,8 @@ summary.weightit.sbps <- function(object, top = 5, ignore.s.weights = FALSE, ...
       attr(w, "focal") <- object$focal
     }
     attr(out, "weights") <- w
-    return(out)
+
+    out
   })
 
   attr(out.list, "prop.subgroup") <- matrix(c(1 - object$prop.subgroup,
@@ -495,6 +612,8 @@ summary.weightit.sbps <- function(object, top = 5, ignore.s.weights = FALSE, ...
   class(out.list) <- "summary.weightit.sbps"
   out.list
 }
+
+#' @exportS3Method print summary.weightit.sbps
 print.summary.weightit.sbps <- function(x, ...) {
   cat("Summary of weights:\n")
   cat("\n - Overall vs. subgroup proportion contribution:\n")
